from typing import Any
from dataclasses import dataclass, fields
from aioquic.buffer import Buffer
from ..types import *
from ..utils.logger import get_logger

logger = get_logger(__name__)

BUF_SIZE = 64


@dataclass
class MOQTMessage:
    """Base class for all MOQT messages."""
    # type: Optional[int] = None - let subclass set it - annoying warnings

    @staticmethod
    def _bytes_encode(value: Any) -> bytes:
        if isinstance(value, int):
            return MOQTMessage._varint_encode(value)
        if isinstance(value, str):
            return value.encode()
        return value

    @staticmethod
    def _varint_encode(value: int) -> bytes:
        buf = Buffer(capacity=8)
        buf.push_uint_var(value)
        return buf.data
    
    @staticmethod
    def _varint_decode(data: bytes) -> int:
        buf = Buffer(data=data)
        return buf.pull_uint_var()

    def serialize(self) -> bytes:
        """Convert message to complete wire format."""
        raise NotImplementedError()

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'MOQTMessage':
        """Create message from buffer containing payload."""
        raise NotImplementedError()

    def __str__(self) -> str:
        """Generic string representation showing all fields."""
        parts = []
        class_fields = fields(self.__class__)

        for field in class_fields:
            value = getattr(self, field.name)
            if "version" in field.name.lower():
                if isinstance(value, (list, tuple)):
                    str_val = "[" + ", ".join(f"0x{x:x}" for x in value) + "]"
                else:
                    str_val = f"0x{value:x}"  # Single version number
            elif field.name == "parameters":
                # Decode parameter types and values
                items = []
                enum = SetupParamType if self.__class__.__name__.endswith('Setup') else ParamType
                for k, v in value.items():
                    param_name = enum(k).name  # Convert enum value to name
                    try:
                        # Try to decode value as varint
                        param_value = v if isinstance(v, int) else self._varint_decode(v)
                        items.append(f"{param_name}={param_value}")
                    except:
                        # Fall back to hex for non-varint values
                        items.append(f"{param_name}=0x{v.hex()}")
                str_val = "{" + ", ".join(items) + "}"
            elif isinstance(value, bytes):
                try:
                    str_val = value.decode('utf-8')
                except UnicodeDecodeError:
                    str_val = f"0x{value.hex()}"
            elif isinstance(value, dict):
                str_val = "{" + ", ".join(f"{k}: {v}" for k, v in value.items()) + "}"
            else:
                str_val = str(value)
            parts.append(f"{field.name}={str_val}")

        return f"{self.__class__.__name__}({', '.join(parts)})"


#!/usr/bin/env python3

import argparse
import logging

import uvloop
import asyncio
from aioquic.h3.connection import H3_ALPN
from aiomoqt.types import ParamType
from aiomoqt.client import MOQTClientSession
from aiomoqt.utils.logger import get_logger, set_log_level


def parse_args():
    parser = argparse.ArgumentParser(description='MOQT WebTransport Client')
    parser.add_argument('--host', type=str, default='localhost', help='Host to connect to')
    parser.add_argument('--port', type=int, default=4433, help='Port to connect to')
    parser.add_argument('--namespace', type=str, default='test', help='Namespace')
    parser.add_argument('--trackname', type=str, default='track', help='Track')
    parser.add_argument('--endpoint', type=str, default='moq', help='MOQT WT endpoint')
    parser.add_argument('--debug', action='store_true',
                        help='Enable debug output')
    return parser.parse_args()


async def main(host: str, port: int, endpoint: str, namespace: str, trackname: str, debug: bool):
    log_level = logging.DEBUG if debug else logging.INFO
    set_log_level(log_level)
    logger = get_logger(__name__)

    client = MOQTClientSession(host, port, endpoint=endpoint, debug=debug)
    logger.info(f"MOQT app: publish session connecting: {client}")
    async with client.connect() as session:
        try: 
            await session.initialize()
            logger.info(f"Publish namespace via ANNOUNCE: {namespace}")
            response = await session.announce(
                namespace=namespace,
                parameters={ParamType.AUTHORIZATION_INFO: b"auth-token-123"},
                wait_response=True,
            )
            logger.info(f"MOQT app: Announce reponse: {response}")
            # process subscription requests - open data streams
            close = await session._moqt_session_close
            logger.info(f"MOQT app: exiting client session: {close}")
        except Exception as e:
            logger.error(f"MOQT session exception: {e}")
            code, reason = session._close if session._close is not None else (0,"Session Closed")
            session.close(error_code=code, reason_phrase=reason)
            pass
    
    logger.info(f"MOQT subscribe session closed: {client}")


if __name__ == "__main__":
    args = parse_args()
    uvloop.run(main(
        host=args.host,
        port=args.port,
        endpoint=args.endpoint,
        namespace=args.namespace,
        trackname=args.trackname,
        debug=args.debug
    ))



#!/usr/bin/env python3
import uvloop
import asyncio
import argparse
import logging

from aioquic.h3.connection import H3_ALPN
from aiomoqt.types import ParamType
from aiomoqt.client import MOQTClientSession
from aiomoqt.messages.announce import SubscribeAnnouncesOk
from aiomoqt.messages.subscribe import SubscribeOk
from aiomoqt.utils.logger import get_logger, set_log_level


def parse_args():
    parser = argparse.ArgumentParser(description='MOQT WebTransport Client')
    parser.add_argument('--host', type=str, default='localhost', help='Host to connect to')
    parser.add_argument('--port', type=int, default=4433, help='Port to connect to')
    parser.add_argument('--namespace', type=str, default="live/test", help='Track Namespace')
    parser.add_argument('--trackname', type=str, default="track", help='Track Name')
    parser.add_argument('--endpoint', type=str, default="moq", help='MOQT WT endpoint')
    parser.add_argument('--debug', action='store_true', help='Enable debug output')
    return parser.parse_args()


async def main(host: str, port: int, endpoint: str, namespace: str, trackname: str, debug: bool):
    log_level = logging.DEBUG if debug else logging.INFO
    set_log_level(log_level)
    logger = get_logger(__name__)

    client = MOQTClientSession(
        host, port, endpoint=endpoint,
        debug=debug
    )
    logger.info(f"MOQT app: subscribe session connecting: {client}")
    async with client.connect() as session:
        try: 
            response = await session.initialize()
            response = await session.subscribe_announces(
                namespace_prefix=namespace,
                parameters={ParamType.AUTHORIZATION_INFO: b"auth-token-123"},
                wait_response=True
            )
            if not isinstance(response, SubscribeAnnouncesOk):
                logger.error(f"SubscribeAnnounces error: {response}")
                raise RuntimeError(response)
            logger.info(f"MOQT app: SubscribeAnnounces response: {response}")
            response = await session.subscribe(
                namespace=namespace,
                track_name=trackname,
                parameters={
                    ParamType.MAX_CACHE_DURATION: 100,
                    ParamType.AUTHORIZATION_INFO: b"auth-token-123",
                    ParamType.DELIVERY_TIMEOUT: 10,
                },
                wait_response=True
            )
            if not isinstance(response, SubscribeOk):
                # logger.error(f"Subscribe error: {response}")
                raise RuntimeError(response)
            # process subscription - publisher will open stream and send data
            close = await session._moqt_session_close
            logger.info(f"MOQT app: exiting client session: {close}")
        except Exception as e:
            logger.error(f"MOQT app: session exception: {e}")
            code, reason = session._close if session._close is not None else (0,"Session Closed")
            session.close(error_code=code, reason_phrase=reason)
            pass
        
    logger.info(f"MOQT app: subscribe session closed: {client}")


if __name__ == "__main__":
    args = parse_args()
    uvloop.run(main(
        host=args.host,
        port=args.port,
        endpoint=args.endpoint,
        namespace=args.namespace,
        trackname=args.trackname,
        debug=args.debug
    ), debug=args.debug)



from ..types import *
from typing import Tuple, Dict, Optional
from dataclasses import dataclass
from aioquic.buffer import Buffer
from .base import MOQTMessage, BUF_SIZE

from ..utils.logger import get_logger

logger = get_logger(__name__)


@dataclass
class TrackStatusRequest(MOQTMessage):
    namespace: Tuple[bytes, ...] = None  # Tuple encoded
    track_name: bytes = None

    def __post_init__(self):
        self.type = MOQTMessageType.TRACK_STATUS_REQUEST

    def serialize(self) -> bytes:
        # Write namespace as tuple
        if not isinstance(self.namespace, tuple):
            raise ValueError("namespace must be a tuple of bytes")
        payload = Buffer()
        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            if not isinstance(part, bytes):
                raise ValueError("namespace parts must be bytes")
            payload.push_uint_var(len(part))
            payload.push_bytes(part)
        
        # Write track name
        if not isinstance(self.track_name, bytes):
            raise ValueError("track_name must be bytes")
        payload.push_uint_var(len(self.track_name))
        payload.push_bytes(self.track_name)

        # Create final message
        buf = Buffer()  # Extra space for header
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload.tell())
        buf.push_bytes(payload.data)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'TrackStatusRequest':
        # Parse namespace tuple
        namespace_parts = []
        num_parts = buffer.pull_uint_var()
        for _ in range(num_parts):
            part_len = buffer.pull_uint_var()
            part = buffer.pull_bytes(part_len)
            namespace_parts.append(part)
        namespace = b'/'.join(namespace_parts)

        # Parse track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        return cls(namespace=namespace, track_name=track_name)


@dataclass
class TrackStatus(MOQTMessage):
    namespace: Tuple[bytes, ...]  # Tuple encoded
    track_name: bytes
    status_code: TrackStatusCode
    last_group_id: int
    last_object_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.TRACK_STATUS

    def serialize(self) -> bytes:
        # First encode the payload
        payload = Buffer(capacity=BUF_SIZE)
        
        # Write namespace as tuple
        if not isinstance(self.namespace, tuple):
            raise ValueError("namespace must be a tuple of bytes")
        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            if not isinstance(part, bytes):
                raise ValueError("namespace parts must be bytes")
            payload.push_uint_var(len(part))
            payload.push_bytes(part)
        
        # Write track name
        if not isinstance(self.track_name, bytes):
            raise ValueError("track_name must be bytes")
        payload.push_uint_var(len(self.track_name))
        payload.push_bytes(self.track_name)

        # Write status info
        if not isinstance(self.status_code, TrackStatusCode):
            raise ValueError("status_code must be TrackStatusCode enum")
        payload.push_uint_var(self.status_code.value)
        payload.push_uint_var(self.last_group_id)
        payload.push_uint_var(self.last_object_id)

        # Create final message
        buf = Buffer(capacity=BUF_SIZE)
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload.tell())
        buf.push_bytes(payload.data)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'TrackStatus':
        # Parse namespace tuple
        namespace_parts = []
        num_parts = buffer.pull_uint_var()
        for _ in range(num_parts):
            part_len = buffer.pull_uint_var()
            part = buffer.pull_bytes(part_len)
            namespace_parts.append(part)
        namespace = b'/'.join(namespace_parts)

        # Parse track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        # Parse status info
        status_code = TrackStatusCode(buffer.pull_uint_var())
        last_group_id = buffer.pull_uint_var()
        last_object_id = buffer.pull_uint_var()

        return cls(
            namespace=namespace,
            track_name=track_name,
            status_code=status_code,
            last_group_id=last_group_id,
            last_object_id=last_object_id
        )


@dataclass
class Subscribe(MOQTMessage):
    """SUBSCRIBE message for requesting track data."""
    subscribe_id: int
    track_alias: int
    namespace: Tuple[bytes, ...]
    track_name: bytes
    priority: int
    direction: int  # Ascending/Descending
    filter_type: int
    start_group: Optional[int] = None
    start_object: Optional[int] = None
    end_group: Optional[int] = None
    parameters: Optional[Dict[int, bytes]] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=BUF_SIZE)
        payload = Buffer(capacity=BUF_SIZE)

        payload.push_uint_var(self.subscribe_id)
        payload.push_uint_var(self.track_alias)

        # Add namespace as tuple
        payload.push_uint_var(len(self.namespace))
        for part in self.namespace:
            payload.push_uint_var(len(part))
            payload.push_bytes(part)

        payload.push_uint_var(len(self.track_name))
        payload.push_bytes(self.track_name)
        payload.push_uint8(self.priority)
        payload.push_uint8(self.direction)
        payload.push_uint_var(self.filter_type)

        # Add optional start/end fields based on filter type
        if self.filter_type in (3, 4):  # ABSOLUTE_START or ABSOLUTE_RANGE
            payload.push_uint_var(self.start_group or 0)
            payload.push_uint_var(self.start_object or 0)

        if self.filter_type == 4:  # ABSOLUTE_RANGE
            payload.push_uint_var(self.end_group or 0)

        # Add parameters
        parameters = self.parameters or {}
        payload.push_uint_var(len(parameters))
        for param_id, param_value in parameters.items():
            payload.push_uint_var(param_id)
            param_value = MOQTMessage._bytes_encode(param_value)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)
        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Subscribe':
        subscribe_id = buffer.pull_uint_var()
        track_alias = buffer.pull_uint_var()

        # Deserialize namespace tuple
        tuple_len = buffer.pull_uint_var()
        namespace = []
        for _ in range(tuple_len):
            part_len = buffer.pull_uint_var()
            namespace.append(buffer.pull_bytes(part_len))
        namespace = tuple(namespace)

        # Track name
        track_name_len = buffer.pull_uint_var()
        track_name = buffer.pull_bytes(track_name_len)

        priority = buffer.pull_uint8()
        direction = buffer.pull_uint8()
        filter_type = buffer.pull_uint_var()

        # Handle optional fields based on filter type
        start_group = None
        start_object = None
        end_group = None
        if filter_type in (3, 4):  # ABSOLUTE_START or ABSOLUTE_RANGE
            start_group = buffer.pull_uint_var()
            start_object = buffer.pull_uint_var()

        if filter_type == 4:  # ABSOLUTE_RANGE
            end_group = buffer.pull_uint_var()

        # Deserialize parameters
        params = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            params[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            track_alias=track_alias,
            namespace=namespace,
            track_name=track_name,
            priority=priority,
            direction=direction,
            filter_type=filter_type,
            start_group=start_group,
            start_object=start_object,
            end_group=end_group,
            parameters=params
        )


@dataclass
class Unsubscribe(MOQTMessage):
    """UNSUBSCRIBE message for ending track subscription."""
    subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.UNSUBSCRIBE

    def serialize(self) -> bytes:
        buf = Buffer(capacity=BUF_SIZE)

        payload = MOQTMessage._varint_encode(self.subscribe_id)

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload))
        buf.push_bytes(payload)
        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'Unsubscribe':
        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)


@dataclass
class SubscribeDone(MOQTMessage):
    """SUBSCRIBE_DONE message indicating subscription completion."""
    subscribe_id: int
    status_code: int  # SubscribeDoneCode
    stream_count: int
    reason: str

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_DONE

    def serialize(self) -> bytes:
        # First encode the payload
        payload = Buffer(capacity=BUF_SIZE)
        
        # Write payload fields
        payload.push_uint_var(self.subscribe_id)
        if not isinstance(self.status_code, SubscribeDoneCode):
            raise ValueError("status_code must be SubscribeDoneCode enum")
        payload.push_uint_var(self.status_code.value)
        payload.push_uint_var(self.stream_count)

        # Convert and write reason string
        if not isinstance(self.reason, str):
            raise ValueError("reason must be str")
        reason_bytes = self.reason.encode()
        payload.push_uint_var(len(reason_bytes))
        payload.push_bytes(reason_bytes)

        # Create final message
        buf = Buffer(capacity=BUF_SIZE)
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload.tell())
        buf.push_bytes(payload.data)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeDone':
        subscribe_id = buffer.pull_uint_var()
        status_code = buffer.pull_uint_var()
        stream_count = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()

        return cls(
            subscribe_id=subscribe_id,
            status_code=status_code,
            stream_count=stream_count,
            reason=reason
        )


@dataclass
class MaxSubscribeId(MOQTMessage):
    """MAX_SUBSCRIBE_ID message setting maximum subscribe ID."""
    subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.MAX_SUBSCRIBE_ID

    def serialize(self) -> bytes:
        buf = Buffer(capacity=BUF_SIZE)

        payload = MOQTMessage._varint_encode(self.subscribe_id)  # subscribe_id varint

        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload))
        buf.push_uint_var(payload)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'MaxSubscribeId':
        subscribe_id = buffer.pull_uint_var()
        return cls(subscribe_id=subscribe_id)


@dataclass
class SubscribesBlocked(MOQTMessage):
    """SUBSCRIBES_BLOCKED message indicating subscriber is blocked."""
    maximum_subscribe_id: int

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBES_BLOCKED

    def serialize(self) -> bytes:
        buf = Buffer(capacity=16)

        payload_size = 1  # maximum_subscribe_id varint

        buf.push_uint_var(self.type)
        buf.push_uint_var(payload_size)
        buf.push_uint_var(self.maximum_subscribe_id)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribesBlocked':
        maximum_subscribe_id = buffer.pull_uint_var()
        return cls(maximum_subscribe_id=maximum_subscribe_id)


@dataclass
class SubscribeOk(MOQTMessage):
    """SUBSCRIBE_OK message indicating successful subscription."""
    subscribe_id: int
    expires: int
    group_order: GroupOrder  # 0x1=Ascending, 0x2=Descending
    content_exists: ContentExistsCode  # 0x0=No Content or 0x1=Exists
    largest_group_id: Optional[int] = None  # Only if content exists
    largest_object_id: Optional[int] = None  # Only if content exists
    parameters: Optional[Dict[int, bytes]] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_OK

    def serialize(self) -> bytes:
        # First encode the payload
        payload = Buffer(capacity=BUF_SIZE)
        
        # Required fields
        payload.push_uint_var(self.subscribe_id)
        payload.push_uint_var(self.expires)
        payload.push_uint8(self.group_order.value)
        payload.push_uint8(self.content_exists)

        # Largest group/object IDs only present if content_exists=1
        if self.content_exists == ContentExistsCode.EXISTS:
            if self.largest_group_id is None or self.largest_object_id is None:
                raise ValueError("largest_group_id and largest_object_id required when content_exists=1")
            payload.push_uint_var(self.largest_group_id)
            payload.push_uint_var(self.largest_object_id)

        # Parameters
        parameters = self.parameters or {}
        payload.push_uint_var(len(parameters))
        for param_type, param_value in parameters.items():
            payload.push_uint_var(param_type)
            param_value = MOQTMessage._bytes_encode(param_value)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        # Create final message
        buf = Buffer(capacity=BUF_SIZE)
        buf.push_uint_var(self.type)
        buf.push_uint_var(len(payload.data))
        buf.push_bytes(payload.data)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeOk':
        subscribe_id = buffer.pull_uint_var()
        expires = buffer.pull_uint_var()
        group_order = GroupOrder(buffer.pull_uint8())
        content_exists = buffer.pull_uint8()

        largest_group_id = None
        largest_object_id = None
        if content_exists == ContentExistsCode.EXISTS:
            largest_group_id = buffer.pull_uint_var()
            largest_object_id = buffer.pull_uint_var()

        parameters = {}
        param_count = buffer.pull_uint_var()
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            parameters[ParamType(param_id)] = param_value

        return cls(
            subscribe_id=subscribe_id,
            expires=expires,
            group_order=group_order,
            content_exists=content_exists,
            largest_group_id=largest_group_id,
            largest_object_id=largest_object_id,
            parameters=parameters
        )

@dataclass
class SubscribeError(MOQTMessage):
    """SUBSCRIBE_ERROR message indicating subscription failure."""
    subscribe_id: int
    error_code: int  # SubscribeErrorCode
    reason: str
    track_alias: int

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_ERROR

    def serialize(self) -> bytes:
        # First encode the payload
        payload = Buffer(capacity=BUF_SIZE)
        
        payload.push_uint_var(self.subscribe_id)
        if not isinstance(self.error_code, SubscribeErrorCode):
            raise ValueError("error_code must be SubscribeErrorCode enum")
        payload.push_uint_var(self.error_code.value)

        # Convert and write reason string
        if not isinstance(self.reason, str):
            raise ValueError("reason must be str")
        reason_bytes = self.reason.encode()
        payload.push_uint_var(len(reason_bytes))
        payload.push_bytes(reason_bytes)
        
        payload.push_uint_var(self.track_alias)

        # Create final message
        buf = Buffer(capacity=BUF_SIZE)
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload.tell())
        buf.push_bytes(payload.data)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeError':
        subscribe_id = buffer.pull_uint_var()
        error_code = buffer.pull_uint_var()
        reason_len = buffer.pull_uint_var()
        reason = buffer.pull_bytes(reason_len).decode()
        track_alias = buffer.pull_uint_var()

        return cls(
            subscribe_id=subscribe_id,
            error_code=error_code,
            reason=reason,
            track_alias=track_alias
        )


@dataclass
class SubscribeUpdate(MOQTMessage):
    """SUBSCRIBE_UPDATE message for modifying an existing subscription."""
    subscribe_id: int
    start_group: int
    start_object: int
    end_group: int
    priority: int
    parameters: Optional[Dict[int, bytes]] = None

    def __post_init__(self):
        self.type = MOQTMessageType.SUBSCRIBE_UPDATE

    def serialize(self) -> bytes:
        # First encode the payload
        payload = Buffer(capacity=BUF_SIZE)
        
        # Write payload fields
        payload.push_uint_var(self.subscribe_id)
        payload.push_uint_var(self.start_group)
        payload.push_uint_var(self.start_object)
        payload.push_uint_var(self.end_group)
        
        if not isinstance(self.priority, int) or not 0 <= self.priority <= 255:
            raise ValueError("priority must be uint8 (0-255)")
        payload.push_uint8(self.priority)

        # Write parameters
        parameters = self.parameters or {}
        payload.push_uint_var(len(parameters))
        for param_id, param_value in parameters.items():
            payload.push_uint_var(param_id.value)
            param_value = MOQTMessage._bytes_encode(param_value)
            payload.push_uint_var(len(param_value))
            payload.push_bytes(param_value)

        # Create final message
        buf = Buffer(capacity=BUF_SIZE)
        buf.push_uint_var(self.type)
        buf.push_uint_var(payload.tell())
        buf.push_bytes(payload.data)

        return buf

    @classmethod
    def deserialize(cls, buffer: Buffer) -> 'SubscribeUpdate':
        subscribe_id = buffer.pull_uint_var()
        start_group = buffer.pull_uint_var()
        start_object = buffer.pull_uint_var()
        end_group = buffer.pull_uint_var()
        priority = buffer.pull_uint8()

        param_count = buffer.pull_uint_var()
        parameters = {}
        for _ in range(param_count):
            param_id = buffer.pull_uint_var()
            param_len = buffer.pull_uint_var()
            param_value = buffer.pull_bytes(param_len)
            parameters[param_id] = param_value

        return cls(
            subscribe_id=subscribe_id,
            start_group=start_group,
            start_object=start_object,
            end_group=end_group,
            priority=priority,
            parameters=parameters
        )



import asyncio
from asyncio import Future
from typing import Optional, Union, Tuple, Dict, Callable
from importlib.metadata import version

from aioquic.buffer import Buffer, UINT_VAR_MAX
from aioquic.asyncio.protocol import QuicConnectionProtocol
from aioquic.quic.connection import QuicConnection, QuicErrorCode
from aioquic.quic.events import QuicEvent, StreamDataReceived
from aioquic.h3.connection import H3Connection, ErrorCode
from aioquic.h3.events import HeadersReceived, DataReceived

from .types import *
from .messages import *
from .utils.logger import get_logger

logger = get_logger(__name__)

USER_AGENT = f"aiomoqt-client/{version('aiomoqt')}"

class MOQTException(Exception):
    def __init__(self, error_code: SessionCloseCode, reason_phrase: str):
        self.error_code = error_code
        self.reason_phrase = reason_phrase
        super().__init__(f"{reason_phrase} ({error_code})")
        

class H3CustomConnection(H3Connection):
    """Custom H3Connection wrapper to support alternate SETTINGS"""
    
    def __init__(self, quic: QuicConnection, table_capacity: int = 0, **kwargs) -> None:
        # settings table capacity can be overridden - this should be generalized
        self._max_table_capacity = table_capacity
        self._max_table_capacity_arg = table_capacity
        super().__init__(quic, **kwargs)
        # report sent settings
        settings = self.sent_settings
        if settings is not None:
            logger.debug("H3 SETTINGS sent:")
            for setting_id, value in settings.items():
                logger.debug(f"  Setting 0x{setting_id:x} = {value}")

    @property
    def _max_table_capacity(self):
        return self._max_table_capacity_arg

    @_max_table_capacity.setter
    def _max_table_capacity(self, value):
        # Ignore the parent class attempt to set it
        pass
    

# base class for client and server session objects
class MOQTSession:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()


class MOQTSessionProtocol(QuicConnectionProtocol):
    """MOQT session protocol implementation."""

    def __init__(self, *args, session: 'MOQTSession', **kwargs):
        super().__init__(*args, **kwargs)
        self._session_cfg = session  # backref to session object with config
        self._h3: Optional[H3Connection] = None
        self._session_id: Optional[int] = None
        self._control_stream_id: Optional[int] = None
        self._streams: Dict[int, Dict] = {}
        self._tasks = set()
        self._loop = asyncio.get_running_loop()
        self._wt_session_fut = self._loop.create_future()
        self._moqt_session_fut = self._loop.create_future()
        self._moqt_session_close = self._loop.create_future()
        self._next_subscribe_id = 1  # prime subscribe id generator
        self._next_track_alias = 1  # prime track alias generator
        self._announce_responses: Dict[int, Future[MOQTMessage]] = {}
        self._subscribe_responses: Dict[int, Future[MOQTMessage]] = {}
        self._unsubscribe_responses: Dict[int, Future[MOQTMessage]] = {}
        self._fetch_responses: Dict[int, Future[MOQTMessage]] = {}
        self._message_registry = dict(MOQTSessionProtocol.MOQT_MESSAGE_REGISTRY)
        self._close = None

    @staticmethod
    def _make_namespace_tuple(namespace: Union[str, Tuple[str, ...]]) -> Tuple[bytes, ...]:
        """Convert string or tuple into bytes tuple."""
        if isinstance(namespace, str):
            return tuple(part.encode() for part in namespace.split('/'))
        elif isinstance(namespace, tuple):
            if all(isinstance(x, bytes) for x in namespace):
                return namespace
            return tuple(part.encode() if isinstance(part, str) else part for part in namespace)
        raise ValueError("namespace must be string with '/' delimiters or tuple")

    def _allocate_subscribe_id(self) -> int:
        """Get next available subscribe ID."""
        subscribe_id = self._next_subscribe_id
        self._next_subscribe_id += 1
        return subscribe_id

    def _allocate_track_alias(self) -> int:
        """Get next available track alias."""
        track_alias = self._next_track_alias
        self._next_track_alias += 1
        return track_alias

    async def initialize(self, timeout: int = 10) -> None:
        """Initialize WebTransport and MOQT session."""
        # Create WebTransport session
        self._session_id = self._h3._quic.get_next_available_stream_id(is_unidirectional=False)

        headers = [
            (b":method", b"CONNECT"),
            (b":protocol", b"webtransport"),
            (b":scheme", b"https"),
            (b":authority",
             f"{self._session_cfg.host}:{self._session_cfg.port}".encode()),
            (b":path", f"/{self._session_cfg.endpoint}".encode()),
            (b"sec-webtransport-http3-draft", b"draft02"),
            (b"user-agent", USER_AGENT.encode()),
        ]

        logger.info(f"H3 send: WebTransport CONNECT: (session id: {self._session_id})")
        for name, value in headers:
            logger.debug(f"  {name.decode()}: {value.decode()}")
            
        self._h3.send_headers(stream_id=self._session_id, headers=headers, end_stream=False)
        self.transmit()
        # Wait for WebTransport session establishment
        try:
            async with asyncio.timeout(timeout):
                result = await self._wt_session_fut
            result = "SUCESS" if result else "FAILED"
            status = "False" if self._close is None else f"True ({self._close})"
            logger.debug(f"H3 event: WebTransport setup: {result} closed: {status}")
        except asyncio.TimeoutError:
            logger.error("WebTransport session establishment timeout")
            raise

        # check for H3 connection close
        if self._close:
            raise RuntimeError(self._close)
    
        # Create MOQT control stream
        self._control_stream_id = self._h3.create_webtransport_stream(session_id=self._session_id)
        logger.info(f"MOQT: Created control stream: (stream id: {self._control_stream_id})")

        # Send CLIENT_SETUP
        client_setup = ClientSetup(
            versions=[0xff000007],
            parameters={SetupParamType.MAX_SUBSCRIBER_ID: MOQTMessage._varint_encode(1000)}
        )
        logger.info(f"MOQT send: {client_setup}")
        self.send_control_message(client_setup.serialize()
        )
        # Wait for SERVER_SETUP
        try:
            result = await self._moqt_session_fut
            logger.info(f"MOQT session setup complete: {result}")
        except asyncio.TimeoutError:
            logger.error("MOQT session setup timeout")
            raise

    # def transmit(self) -> None:
    #     """Transmit pending data."""
    #     logger.debug("Transmitting data")
    #     super().transmit()

    def connection_made(self, transport):
        """Called when QUIC connection is established."""
        super().connection_made(transport)
        self._h3 = H3CustomConnection(self._quic, enable_webtransport=True)
        logger.info("H3 connection initialized")

    def quic_event_received(self, event: QuicEvent) -> None:
        """Handle incoming QUIC events."""
        # Log any errors
        if hasattr(event, 'error_code'):
            error = getattr(event, 'error_code') if hasattr(event, 'error_code') else QuicErrorCode.INTERNAL_ERROR
            reason = getattr(event, 'reason_phrase') if hasattr(event, 'reason_phrase') else event.__class__.__name__
            logger.error(f"QUIC error: {error} reason: {reason}")
            if not self._moqt_session_close.done():
                self._moqt_session_close.set_result((error,reason))

        stream_id = getattr(event, 'stream_id', 'unknown')
        classname = event.__class__.__name__
        data = "0x" + event.data.hex() if (hasattr(event, 'data') and event.data is not None) else ""
        logger.debug(f"QUIC event: stream {stream_id}: {classname}: {data}")

        # Handle MOQT control and data messages
        if isinstance(event, StreamDataReceived):
            if stream_id == self._control_stream_id:
                msgbuf = Buffer(data=event.data)
                while msgbuf.tell() < msgbuf.capacity:
                    msg = self.handle_control_message(msgbuf)
                return
            elif stream_id in self._streams:
                self.handle_data_message(stream_id, event.data)
                return

        # Pass remaining events to H3
        if self._h3 is not None:
            try:
                for h3_event in self._h3.handle_event(event):
                    self._h3_event_received(h3_event)

                if (isinstance(event, StreamDataReceived) and hasattr(event, 'data') and
                    event.data is not None and event.data[:2] == b'\x00\x04'):
                    logger.debug("H3 SETTINGS received:")
                    # Check received settings
                    settings = self._h3.received_settings
                    if settings is not None:
                        for setting_id, value in settings.items():
                            logger.debug(f"  Setting 0x{setting_id:x} = {value}")

            except Exception as e:
                logger.error(f"H3 error: error handling event: {e}")
                raise
        else:
            logger.error(f"QUIC event: stream {stream_id}: event not handled({classname})")

    def handle_control_message(self, buffer: Buffer) -> Optional[MOQTMessage]:
        """Process an incoming message."""
        if buffer.capacity <= 0:
            logger.warning("handle_control_message: no data")
            return None

        try:
            start_pos = buffer.tell()
            msg_type = buffer.pull_uint_var()
            msg_len = buffer.pull_uint_var()
            hdr_len = buffer.tell() - start_pos
            # Look up message class
            message_class, handler = self._message_registry.get(msg_type)
            if message_class is None:
                logger.error(f"MOQT error: unknown control message ({hex(msg_type)})")
                return None

            # Deserialize message
            msg = message_class.deserialize(buffer)
            msg_len += hdr_len
            assert (start_pos + msg_len == buffer.tell())
            logger.info(f"MOQT event: control message: {msg} ({msg_len} bytes)")

            # Schedule handler if one exists
            if handler:
                task = asyncio.create_task(handler(self, msg))
                task.add_done_callback(lambda t: self._tasks.discard(t))
                self._tasks.add(task)


            return msg

        except Exception as e:
            logger.error(f"handle_control_message: error handling control message: {e}")
            raise

    def handle_data_message(self, stream_id: int, buffer: Buffer) -> None:
        """Process incoming data messages (not control messages)."""
        if buffer.capacity <= 0:
            logger.error(f"MOQT event: stream {stream_id}: message contains no data")
            return

        try:
            # Get stream type from first byte
            stream_type = buffer.pull_uint_var()
            logger.info(f"MOQT: stream {stream_id}: type: {hex(stream_type)}")

            # Look up stream message class
            message_class = self._stream_types.get(stream_type)
            if message_class is None:
                logger.error(f"MOQT error: unknown stream type ({hex(stream_type)})")
                return

            if stream_type == StreamType.STREAM_HEADER_SUBGROUP:
                msg = message_class.deserialize(buffer)
                logger.info(f"MOQT event: stream message: {msg}")
                self._handle_subgroup_header(msg)
                return

            # Process remaining buffer as object data
            self._handle_object_data(buffer)

        except Exception as e:
            logger.error(f"handle_data_message: error handling data message: {e}")


    def _h3_event_received(self, event: QuicEvent) -> None:
        """Handle H3-specific events."""
        if isinstance(event, HeadersReceived):
            self._h3_handle_headers_received(event)
        elif isinstance(event, DataReceived):
            self._h3_handle_data_received(event)
        else:
            msg_class = event.__class__.__name__
            data = "0x" + event.data.hex() if (hasattr(event, 'data') and event.data is not None) else ""
            logger.error(f"H3 error: stream {event.stream_id}: {msg_class}: {data}")

    def _h3_handle_headers_received(self, event: HeadersReceived) -> None:
        """Process incoming H3 headers."""
        status = None
        logger.info(f"H3 event: HeadersReceived: (session id: {event.stream_id})")
        for name, value in event.headers:
            logger.debug(f"  {name.decode()}: {value.decode()}")
            if name == b':status':
                status = value

        stream_id = event.stream_id
        if status == b"200":
            logger.info(f"H3 event: WebTransport session established: (session id: {stream_id})")
            self._wt_session_fut.set_result(True)
        else:
            error = f"WebTransport session setup failed ({status})"
            logger.error(f"H3 error: stream {stream_id}: " + error)
            self._close = (ErrorCode.H3_CONNECT_ERROR, error)
            self._wt_session_fut.set_result(False)

    def _h3_handle_data_received(self, event: DataReceived) -> None:
        """Process incoming H3 data MOQT data is not expected to arrive like this"""
        logger.warning(f"H3 event: stream {event.stream_id}: DataReceived")
        if hasattr(event, 'data'):
            logger.debug(f"H3 event: stream {event.stream_id}: data: 0x{event.data.hex()}")

    def register_handler(self, msg_type: int, handler: Callable) -> None:
        """Register a custom message handler."""
        self._custom_handlers[msg_type] = handler

    def send_control_message(self, buf: Buffer) -> None:
        """Send a MOQT message on the control stream."""
        if self._quic is None or self._control_stream_id is None:
            raise RuntimeError("Control stream not initialized")

        self._quic.send_stream_data(
            stream_id=self._control_stream_id,
            data=buf.data,
            end_stream=False
        )
        self.transmit()

    def close(self, error_code: SessionCloseCode = SessionCloseCode.NO_ERROR, reason_phrase: str = "MOQT Session Close") -> None:
        """Close the MOQT session."""
        logger.info(f"MOQT session: Closing: {reason_phrase} ({error_code})")
        if self._session_id is not None:
            logger.debug(f"H3 session: Closing: stream {self._session_id} {self._h3}")
            self._h3.send_data(self._session_id, b"", end_stream=True)
            self._session_id = None
        self._h3 = None
        # set the exit condition for the async with session
        if not self._moqt_session_close.done:
            self._moqt_session_close.set_result(True)
        # call parent close and transmit all
        super().close(error_code=error_code, reason_phrase=reason_phrase)
        self.transmit()


    ################################################################################################
    #  Outbound control message API - note: awaitable messages support 'wait_response' param       #
    ################################################################################################
    def subscribe(
        self,
        namespace: str,
        track_name: str,
        subscribe_id: int = None,
        track_alias: int = None,
        priority: int = 128,
        group_order: GroupOrder = GroupOrder.ASCENDING,
        filter_type: FilterType = FilterType.LATEST_GROUP,
        start_group: Optional[int] = 0,
        start_object: Optional[int] = 0,
        end_group: Optional[int] = 0,
        parameters: Optional[Dict[int, bytes]] = None,
        wait_response: Optional[bool] = False,
    ) -> None:
        """Subscribe to a track with configurable options."""
        if parameters is None:
            parameters = {}
        subscribe_id = self._allocate_subscribe_id() if subscribe_id is None else subscribe_id
        track_alias = self._allocate_track_alias() if track_alias is None else track_alias
        namespace_tuple = self._make_namespace_tuple(namespace)
        message = Subscribe(
            subscribe_id=subscribe_id,
            track_alias=track_alias,
            namespace=namespace_tuple,
            track_name=track_name.encode(),
            priority=priority,
            direction=group_order,
            filter_type=filter_type,
            start_group=start_group,
            start_object=start_object,
            end_group=end_group,
            parameters=parameters
        )
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())

        if not wait_response:
            return message

        # Create future for response
        subscribe_fut = self._loop.create_future()
        self._subscribe_responses[subscribe_id] = subscribe_fut

        async def wait_for_response():
            try:
                async with asyncio.timeout(10):
                    response = await subscribe_fut
            except asyncio.TimeoutError:
                # Create synthetic error response
                response = SubscribeError(
                    subscribe_id=subscribe_id,
                    error_code=0x5,  # TIMEOUT error code
                    reason="Subscribe Response Timeout",
                    track_alias=0
                )
                logger.error(f"Timeout waiting for subscribe response")
            finally:
                logger.debug(f"MOQT: removing subscribe response future: {subscribe_id}")
                self._subscribe_responses.pop(subscribe_id, None)    
            return response

        return wait_for_response()

    def subscribe_ok(
        self,
        subscribe_id: int,
        expires: int = 0,  # 0 means no expiry
        group_order: int = GroupOrder.ASCENDING,
        content_exists: int = 0,
        largest_group_id: Optional[int] = None,
        largest_object_id: Optional[int] = None,
        parameters: Optional[Dict[int, bytes]] = None
    ) -> SubscribeOk:
        """Create and send a SUBSCRIBE_OK response."""
        message = SubscribeOk(
            subscribe_id=subscribe_id,
            expires=expires,
            group_order=group_order,
            content_exists=content_exists,
            largest_group_id=largest_group_id,
            largest_object_id=largest_object_id,
            parameters=parameters or {}
        )
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())
        return message

    def subscribe_error(
        self,
        subscribe_id: int,
        error_code: int = SubscribeErrorCode.INTERNAL_ERROR,
        reason: str = "Internal error",
        track_alias: Optional[int] = None
    ) -> SubscribeError:
        """Create and send a SUBSCRIBE_ERROR response."""
        message = SubscribeError(
            subscribe_id=subscribe_id,
            error_code=error_code,
            reason=reason,
            track_alias=track_alias
        )
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())
        return message

    def announce(
        self,
        namespace: Union[str, Tuple[str, ...]],
        parameters: Optional[Dict[int, bytes]] = None,
        wait_response: Optional[bool] = False
    ) -> Announce:
        """Announce track namespace availability."""
        namespace_tuple = self._make_namespace_tuple(namespace)
        message = Announce(
            namespace=namespace_tuple,
            parameters=parameters or {}
        )
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())

        if not wait_response:
            return message

        # Create future for response
        announce_fut = self._loop.create_future()
        self._announce_responses[namespace_tuple] = announce_fut

        async def wait_for_response():
            try:
                async with asyncio.timeout(10):
                    response = await announce_fut
            except asyncio.TimeoutError:
                # Create synthetic error response
                response = AnnounceError(
                    namespace=namespace,
                    error_code=0x5,  # TIMEOUT error code
                    reason="Response timeout"
                )
                logger.error(f"Timeout waiting for announce response")
            finally:
                self._announce_responses.pop(namespace, None)
            return response

        return wait_for_response()

    def announce_ok(
        self,
        namespace: Union[str, Tuple[str, ...]],
    ) -> AnnounceOk:
        """Create and send a ANNOUNCE_OK response."""
        namespace_tuple = self._make_namespace_tuple(namespace)
        message = AnnounceOk(
            namespace=namespace_tuple,
        )
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())
        return message

    def unannounce(
        self,
        namespace: Tuple[bytes, ...]
    ) -> None:
        """Withdraw track namespace announcement. (no reply expected)"""        
        message =  Unannounce(namespace=namespace)
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())
        return message
    
    def unsubscribe(
        self,
        subscribe_id: int,
        wait_response: Optional[bool] = False
    ) -> None:
        """Unsubscribe from a track."""
        message = Unsubscribe(subscribe_id=subscribe_id)
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())
 
        if not wait_response:
            return message       

        # Create future for response
        unsubscribe_fut = self._loop.create_future()
        self._unsubscribe_responses[subscribe_id] = unsubscribe_fut

        async def wait_for_response():
            try:
                async with asyncio.timeout(10):
                    response = await unsubscribe_fut
            except asyncio.TimeoutError:
                # Create synthetic error response
                response = SubscribeDone(
                    subscribe_id=subscribe_id,
                    stream_count=0,  # XXX what to do
                    status_code=SubscribeDoneCode.INTERNAL_ERROR,  # TIMEOUT error code
                    reason="Unsubscribe Response Timeout"
                )
                logger.error(f"Timeout waiting for announce response")
            finally:
                self._unsubscribe_responses.pop(subscribe_id, None)
            return response

        return wait_for_response()

    def subscribe_announces(
        self,
        namespace_prefix: str,
        parameters: Optional[Dict[int, bytes]] = None,
        wait_response: Optional[bool] = False
    ) -> None:
        """Subscribe to announcements for a namespace prefix."""
        if parameters is None:
            parameters = {}

        prefix = self._make_namespace_tuple(namespace_prefix)
        message = SubscribeAnnounces(
            namespace_prefix=prefix,
            parameters=parameters
        )
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())

        if not wait_response:
            return message

        # Create future for response
        sub_announces_fut = self._loop.create_future()
        self._subscribe_responses[prefix] = sub_announces_fut

        async def wait_for_response():
            try:
                async with asyncio.timeout(10):
                    response = await sub_announces_fut
            except asyncio.TimeoutError:
                # Create synthetic error response
                response = SubscribeAnnouncesError(
                    namespace_prefix=prefix,
                    error_code=0x5,  # TIMEOUT error code
                    reason="Response timeout"
                )
                logger.error(f"Timeout waiting for announce subscribe response")
            finally:
                self._subscribe_responses.pop(prefix, None)
            
            return response

        return wait_for_response()

    def subscribe_announces_ok(
        self,
        namespace_prefix: Union[str, Tuple[str, ...]],
    ) -> SubscribeAnnouncesOk:
        """Create and send a SUBSCRIBE_ANNOUNCES_OK response."""
        namespace_tuple = self._make_namespace_tuple(namespace_prefix)
        logger.info(f"MOQT send: {message}")
        message = SubscribeAnnouncesOk(namespace_prefix=namespace_tuple)
        self.send_control_message(message.serialize())
        return message

    def unsubscribe_announces(
        self,
        namespace_prefix: str
    ) -> None:
        """Unsubscribe from announcements for a namespace prefix."""        
        prefix = self._make_namespace_tuple(namespace_prefix)
        message = UnsubscribeAnnounces(namespace_prefix=prefix)
        logger.info(f"MOQT send: {message}")
        self.send_control_message(message.serialize())
        return message


    ###############################################################################################
    #  Inbound MoQT message handlers                                                              #
    ###############################################################################################
    async def _handle_server_setup(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, ServerSetup)
        logger.info(f"MOQT handle: {msg}")

        if self._moqt_session_fut.done():
            error = "Received multiple SERVER_SETUP message"
            logger.error(error)
            self.close(
                error_code=SessionCloseCode.PROTOCOL_VIOLATION,
                reason_phrase=error
            )
            raise RuntimeError(error)
        # indicate moqt session setup is complete
        self._moqt_session_fut.set_result(True)

    async def _handle_client_setup(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, ClientSetup)
        logger.info(f"MOQT handle: {msg}")
        # Send SERVER_SETUP in response

    async def _handle_subscribe(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, Subscribe)
        logger.info(f"MOQT receive: {msg}")
        self.subscribe_ok(msg.subscribe_id)

    async def _handle_announce(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, Announce)
        logger.info(f"MOQT receive: {msg}")
        self.announce_ok(msg.namespace)

    async def _handle_subscribe_update(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeUpdate)
        logger.info(f"MOQT handle: {msg}")
        # Handle subscription update

    async def _handle_subscribe_ok(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeOk)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for subscriber waiting for response
        future = self._subscribe_responses.get(msg.subscribe_id)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_subscribe_error(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeError)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for subscriber waiting for response
        future = self._subscribe_responses.get(msg.subscribe_id)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_announce_ok(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, AnnounceOk)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for announcer waiting for response
        future = self._announce_responses.get(msg.namespace)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_announce_error(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, AnnounceError)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for announcer waiting for response
        future = self._announce_responses.get(msg.namespace)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_unannounce(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, Unannounce)
        logger.info(f"MOQT handle: {msg}")
        self.announce_ok(msg.namespace)

    async def _handle_announce_cancel(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, AnnounceCancel)
        logger.info(f"MOQT handle: {msg}")
        # Handle announcement cancellation

    async def _handle_unsubscribe(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, Unsubscribe)
        logger.info(f"MOQT handle: {msg}")
        # Handle unsubscribe request

    async def _handle_subscribe_done(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeDone)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for subscriber waiting for completion
        future = self._subscribe_responses.get(msg.subscribe_id)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_max_subscribe_id(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, MaxSubscribeId)
        logger.info(f"MOQT handle: {msg}")
        # Update maximum subscribe ID

    async def _handle_subscribes_blocked(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribesBlocked)
        logger.info(f"MOQT handle: {msg}")
        # Handle subscribes blocked notification

    async def _handle_track_status_request(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, TrackStatusRequest)
        logger.info(f"MOQT handle: {msg}")
        # Send track status in response

    async def _handle_track_status(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, TrackStatus)
        logger.info(f"MOQT handle: {msg}")
        # Handle track status update

    async def _handle_goaway(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, GoAway)
        logger.info(f"MOQT handle: {msg}")
        # Handle session migration request

    async def _handle_subscribe_announces(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeAnnounces)
        logger.info(f"MOQT handle: {msg}")
        self.subscribe_announces_ok(msg.namespace_prefix)
           
    async def _handle_subscribe_announces_ok(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeAnnouncesOk)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for subscriber waiting for response
        future = self._subscribe_responses.get(msg.namespace_prefix)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_subscribe_announces_error(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, SubscribeAnnouncesError)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for subscriber waiting for response
        future = self._subscribe_responses.get(msg.namespace_prefix)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_unsubscribe_announces(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, UnsubscribeAnnounces)
        logger.info(f"MOQT handle: {msg}")
        self.subscribe_announces_ok(msg.namespace_prefix)

    async def _handle_fetch(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, Fetch)
        logger.info(f"MOQT handle: {msg}")
        self.fetch_ok(msg.subscribe_id)

    async def _handle_fetch_cancel(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, FetchCancel)
        logger.info(f"MOQT handle: {msg}")
        # Handle fetch cancellation

    async def _handle_fetch_ok(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, FetchOk)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for fetcher waiting for response
        future = self._fetch_responses.get(msg.subscribe_id)
        if future and not future.done():
            future.set_result(msg)

    async def _handle_fetch_error(self, msg: MOQTMessage) -> None:
        assert isinstance(msg, FetchError)
        logger.info(f"MOQT handle: {msg}")
        # Set future result for fetcher waiting for response
        future = self._fetch_responses.get(msg.subscribe_id)
        if future and not future.done():
            future.set_result(msg)

    # MoQT message classes for serialize/deserialize, message handler methods (unbound)       
    MOQT_MESSAGE_REGISTRY = {
       # Setup messages
       MOQTMessageType.CLIENT_SETUP: (ClientSetup, _handle_client_setup),
       MOQTMessageType.SERVER_SETUP: (ServerSetup, _handle_server_setup),

       # Subscribe messages
       MOQTMessageType.SUBSCRIBE_UPDATE: (SubscribeUpdate, _handle_subscribe_update),
       MOQTMessageType.SUBSCRIBE: (Subscribe, _handle_subscribe),
       MOQTMessageType.SUBSCRIBE_OK: (SubscribeOk, _handle_subscribe_ok), 
       MOQTMessageType.SUBSCRIBE_ERROR: (SubscribeError, _handle_subscribe_error),

       # Announce messages
       MOQTMessageType.ANNOUNCE: (Announce, _handle_announce),
       MOQTMessageType.ANNOUNCE_OK: (AnnounceOk, _handle_announce_ok),
       MOQTMessageType.ANNOUNCE_ERROR: (AnnounceError, _handle_announce_error),
       MOQTMessageType.UNANNOUNCE: (Unannounce, _handle_unannounce),
       MOQTMessageType.ANNOUNCE_CANCEL: (AnnounceCancel, _handle_announce_cancel),

       # Subscribe control messages
       MOQTMessageType.UNSUBSCRIBE: (Unsubscribe, _handle_unsubscribe),
       MOQTMessageType.SUBSCRIBE_DONE: (SubscribeDone, _handle_subscribe_done),
       MOQTMessageType.MAX_SUBSCRIBE_ID: (MaxSubscribeId, _handle_max_subscribe_id),
       MOQTMessageType.SUBSCRIBES_BLOCKED: (SubscribesBlocked, _handle_subscribes_blocked),

       # Status messages
       MOQTMessageType.TRACK_STATUS_REQUEST: (TrackStatusRequest, _handle_track_status_request),
       MOQTMessageType.TRACK_STATUS: (TrackStatus, _handle_track_status),

       # Session control messages
       MOQTMessageType.GOAWAY: (GoAway, _handle_goaway),

       # Subscribe announces messages
       MOQTMessageType.SUBSCRIBE_ANNOUNCES: (SubscribeAnnounces, _handle_subscribe_announces),
       MOQTMessageType.SUBSCRIBE_ANNOUNCES_OK: (SubscribeAnnouncesOk, _handle_subscribe_announces_ok),
       MOQTMessageType.SUBSCRIBE_ANNOUNCES_ERROR: (SubscribeAnnouncesError, _handle_subscribe_announces_error),
       MOQTMessageType.UNSUBSCRIBE_ANNOUNCES: (UnsubscribeAnnounces, _handle_unsubscribe_announces),

       # Fetch messages
       MOQTMessageType.FETCH: (Fetch, _handle_fetch),
       MOQTMessageType.FETCH_CANCEL: (FetchCancel, _handle_fetch_cancel),
       MOQTMessageType.FETCH_OK: (FetchOk, _handle_fetch_ok),
       MOQTMessageType.FETCH_ERROR: (FetchError, _handle_fetch_error),
   }


import pytest
from aiomoqt.protocol import MOQTSessionProtocol
from aiomoqt.client import MOQTClientSession
from aiomoqt.types import *
from aiomoqt.messages import *


def test_moqt_stub():
    assert True


import ssl
from typing import Optional, AsyncContextManager

from aioquic.quic.configuration import QuicConfiguration
from aioquic.asyncio.client import connect
from aioquic.h3.connection import H3_ALPN

from .protocol import MOQTSession, MOQTSessionProtocol
from .utils.logger import get_logger, QuicDebugLogger

logger = get_logger(__name__)

class MOQTClientSession(MOQTSession):  # New connection manager class
    def __init__(
        self,
        host: str,
        port: int,
        endpoint: Optional[str] = None,
        configuration: Optional[QuicConfiguration] = None,
        debug: bool = False
    ):
        self.host = host
        self.port = port
        self.debug = debug
        self.endpoint = endpoint
        if configuration is None:
            configuration = QuicConfiguration(
                alpn_protocols=H3_ALPN,
                is_client=True,
                verify_mode=ssl.CERT_NONE,
                quic_logger=QuicDebugLogger() if debug else None,
                secrets_log_file=open(
                    "/tmp/keylog.client.txt", "a") if debug else None
            )
        self.configuration = configuration
        logger.debug(f"quic_logger: {configuration.quic_logger.__class__}")

    def connect(self) -> AsyncContextManager[MOQTSessionProtocol]:
        """Return a context manager that creates MOQTSessionProtocol instance."""
        logger.debug(f"MOQT: session connect: {self}")
        protocol = lambda *args, **kwargs: MOQTSessionProtocol(*args, **kwargs, session=self)
        return connect(
            self.host,
            self.port,
            configuration=self.configuration,
            create_protocol=protocol
        )



from ..types import MOQTMessageType
from .base import MOQTMessage
from .setup import ClientSetup, ServerSetup, GoAway
from .announce import (
    Announce, AnnounceOk, AnnounceError, Unannounce, AnnounceCancel,
    SubscribeAnnounces, SubscribeAnnouncesOk, SubscribeAnnouncesError, UnsubscribeAnnounces
)
from .subscribe import (
    Subscribe, SubscribeOk, SubscribeError, SubscribeUpdate, Unsubscribe, SubscribeDone,
    MaxSubscribeId, SubscribesBlocked, TrackStatusRequest, TrackStatus
)
from .fetch import Fetch, FetchOk, FetchError, FetchCancel
from .trackdata import StreamHeaderSubgroup, FetchHeader, ObjectDatagram, ObjectDatagramStatus

__all__ = [
    'MOQTMessage', 'MOQTMessageType',
    'ClientSetup', 'ServerSetup', 'GoAway',
    'Subscribe', 'SubscribeOk', 'SubscribeError', 'SubscribeUpdate',
    'Unsubscribe', 'SubscribeDone', 'MaxSubscribeId', 'SubscribesBlocked',
    'TrackStatusRequest', 'TrackStatus',
    'Announce', 'AnnounceOk', 'AnnounceError', 'Unannounce', 'AnnounceCancel',
    'SubscribeAnnounces', 'SubscribeAnnouncesOk', 'SubscribeAnnouncesError',
    'UnsubscribeAnnounces',
    'Fetch', 'FetchOk', 'FetchError', 'FetchCancel',
    'StreamHeaderSubgroup', 'FetchHeader',
    'ObjectDatagram', 'ObjectDatagramStatus'
    ]
